from __future__ import annotations

import ruamel.yaml

from .base import Transform


class GitLabReferenceExpectationViolation(ValueError):
    pass


class GitLabReference:
    yaml_tag = "!reference"

    @classmethod
    def from_yaml(
        cls, constructor: ruamel.yaml.BaseConstructor, node: ruamel.yaml.Node
    ) -> list[str]:
        if not isinstance(node.value, list):
            raise GitLabReferenceExpectationViolation(
                "check-jsonschema rejects this gitlab !reference tag: "
                f"non-list-value\n{node!r}"
            )
        return [item.value for item in node.value]


# this "transform" is actually a no-op on the data, but it registers the GitLab !reference
# tag with the instance YAML loader
class GitLabDataTransform(Transform):
    def modify_yaml_implementation(self, implementation: ruamel.yaml.YAML) -> None:
        implementation.register_class(GitLabReference)


GITLAB_TRANSFORM = GitLabDataTransform()



from __future__ import annotations

import typing as t
import urllib.parse


def _bitbucket_pipelines_url() -> str:
    return "https://api.bitbucket.org/schemas/pipelines-configuration"


def _githubusercontent_url(owner: str, repo: str, ref: str, path: str) -> str:
    return f"https://raw.githubusercontent.com/{owner}/{repo}/{ref}/{path}"


def _gitlab_raw_url(repo: str, file_path: str, ref: str) -> str:
    return (
        "https://gitlab.com/api/v4"
        f"/projects/{urllib.parse.quote(repo, safe='')}/repository/files"
        f"/{urllib.parse.quote(file_path, safe='')}/raw?ref={ref}"
    )


# this lists custom schemas which are *not* part of the vendored schema catalog
CUSTOM_SCHEMA_CATALOG: dict[str, dict[str, t.Any]] = {
    "github-workflows-require-timeout": {
        "hook_config": {
            "name": "Require timeout-minutes in GitHub Workflows",
            "description": (
                "Validate that all GitHub Workflow jobs set timeout-minutes"
            ),
            "files": r"^\.github/workflows/[^/]+$",
            "types": "yaml",
        },
    },
}
CUSTOM_SCHEMA_NAMES = list(CUSTOM_SCHEMA_CATALOG)

# Known configs. The SchemaCatalog lists known schema URLs with their names.
# kept in alphabetical order by name
#
# Additional config could be associated with the schemas in the future.
SCHEMA_CATALOG: dict[str, dict[str, t.Any]] = {
    "azure-pipelines": {
        "url": _githubusercontent_url(
            "microsoft", "azure-pipelines-vscode", "main", "service-schema.json"
        ),
        "hook_config": {
            "name": "Validate Azure Pipelines",
            "description": (
                "Validate Azure Pipelines config against the schema provided "
                "by Microsoft"
            ),
            "add_args": [
                "--data-transform",
                "azure-pipelines",
                "--regex-variant",
                "nonunicode",
            ],
            "files": r"^(\.)?azure-pipelines\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "bamboo-spec": {
        "url": "https://json.schemastore.org/bamboo-spec.json",
        "hook_config": {
            "name": "Validate Bamboo Specs",
            "files": r"^bamboo-specs/.*\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "bitbucket-pipelines": {
        "url": _bitbucket_pipelines_url(),
        "hook_config": {
            "name": "Validate Bitbucket Pipelines",
            "files": r"bitbucket-pipelines\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "buildkite": {
        "url": _githubusercontent_url(
            "buildkite", "pipeline-schema", "main", "schema.json"
        ),
        "hook_config": {
            "name": "Validate Buildkite Pipelines",
            "description": (
                "Validate Buildkite Pipelines against the schema provided by Buildkite"
            ),
            "files": [
                r"buildkite\.(yml|yaml|json)",
                r"buildkite\.(.+)\.(yml|yaml|json)",
                r"(.*/)?\.buildkite/pipeline\.(yml|yaml|json)",
                r"(.*/)?\.buildkite/pipeline\.(.+)\.(yml|yaml|json)",
            ],
            "types_or": ["json", "yaml"],
        },
    },
    "changie": {
        "url": "https://changie.dev/schema.json",
        "hook_config": {
            "name": "Validate Changie config",
            "description": "Validate Changie configuration",
            "files": r"^\.changie\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "circle-ci": {
        "url": _githubusercontent_url(
            "CircleCI-Public", "circleci-yaml-language-server", "main", "schema.json"
        ),
        "hook_config": {
            "name": "Validate CircleCI config",
            "description": (
                "Validate CircleCI config against the schema provided by SchemaStore"
            ),
            "files": r"^\.circleci/config\.(yml|yaml)$",
            "type": "yaml",
        },
    },
    "citation-file-format": {
        "url": _githubusercontent_url(
            "citation-file-format",
            "citation-file-format",
            "main",
            "schema.json",
        ),
        "hook_config": {
            "name": "Validate Citation File Format",
            "description": "Validate Citation File Format",
            "files": r"^CITATION.cff$",
        },
    },
    "cloudbuild": {
        "url": "https://json.schemastore.org/cloudbuild.json",
        "hook_config": {
            "name": "Validate Google Cloud Build config",
            "description": (
                "Validate Google Cloud Build config against the schema provided "
                "by SchemaStore"
            ),
            "files": r"^cloudbuild\.(yml|yaml|json)$",
            "types_or": ["json", "yaml"],
        },
    },
    "codecov": {
        "url": "https://www.schemastore.org/codecov.json",
        "hook_config": {
            "name": "Validate Codecov config",
            "files": [
                r"^((\.github|dev)/)?\.?codecov\.ya?ml$",
            ],
            "types": "yaml",
        },
    },
    "compose-spec": {
        "url": _githubusercontent_url(
            "compose-spec",
            "compose-spec",
            "master",
            "schema/compose-spec.json",
        ),
        "hook_config": {
            "name": "Validate Docker Compose files",
            "files": [
                r"([^/]*/)*docker-compose(\.[\.a-zA-Z0-9_-]*)*\.(yml|yaml)",
                r"([^/]*/)*compose(\.[\.a-zA-Z0-9_-]*)*\.(yml|yaml)",
            ],
            "types": "yaml",
        },
    },
    "dependabot": {
        "url": "https://json.schemastore.org/dependabot-2.0.json",
        "hook_config": {
            "name": "Validate Dependabot Config (v2)",
            "files": r"^\.github/dependabot\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "drone-ci": {
        "url": "https://json.schemastore.org/drone.json",
        "hook_config": {
            "name": "Validate Drone-CI Config",
            "files": r"^\.drone\.yml$",
            "types": "yaml",
        },
    },
    "github-actions": {
        "url": "https://json.schemastore.org/github-action",
        "hook_config": {
            "name": "Validate GitHub Actions",
            "files": [
                r"action\.(yml|yaml)",
                r"\.github/actions/(.+/)?action\.(yml|yaml)",
            ],
            "types": "yaml",
        },
    },
    "github-discussion": {
        "url": "https://www.schemastore.org/github-discussion.json",
        "hook_config": {
            "name": "Validate GitHub discussion",
            "files": [
                r"^\.github/DISCUSSION_TEMPLATE/.+\.ya?ml$",
            ],
            "types": "yaml",
        },
    },
    "github-issue-config": {
        "url": "https://www.schemastore.org/github-issue-config.json",
        "hook_config": {
            "name": "Validate GitHub issue config",
            "files": [
                r"^\.github/ISSUE_TEMPLATE/config\.yml$",
            ],
            "types": "yaml",
        },
    },
    "github-issue-forms": {
        "url": "https://www.schemastore.org/github-issue-forms.json",
        "hook_config": {
            "name": "Validate GitHub issue forms",
            "files": [
                r"^\.github/ISSUE_TEMPLATE/(?!config\.ya?ml$).+$",
            ],
            "types": "yaml",
        },
    },
    "github-workflows": {
        "url": "https://json.schemastore.org/github-workflow",
        "hook_config": {
            "name": "Validate GitHub Workflows",
            "files": r"^\.github/workflows/[^/]+$",
            "types": "yaml",
        },
    },
    "gitlab-ci": {
        "url": _gitlab_raw_url(
            "gitlab-org/gitlab",
            "app/assets/javascripts/editor/schema/ci.json",
            "master",
        ),
        "hook_config": {
            "name": "Validate GitLab CI config",
            "add_args": [
                "--data-transform",
                "gitlab-ci",
                "--regex-variant",
                "nonunicode",
            ],
            "files": r"^.*\.gitlab-ci\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "meltano": {
        "url": _githubusercontent_url(
            "meltano",
            "meltano",
            "main",
            "src/meltano/schemas/meltano.schema.json",
        ),
        "hook_config": {
            "name": "Validate Meltano Config",
            "description": (
                "Validate Meltano config against the schema provided by Meltano"
            ),
            "files": [
                r".*meltano\.yml",
                r"meltano-manifest\.json",
                r"meltano-manifest\..+\.json",
            ],
            "types_or": ["json", "yaml"],
        },
    },
    "mergify": {
        "url": "https://docs.mergify.com/mergify-configuration-schema.json",
        "hook_config": {
            "name": "Validate Mergify config",
            "files": [
                r"\.mergify\.yml",
                r"\.mergify/config\.yml",
                r"\.github/mergify\.yml",
            ],
            "types": "yaml",
        },
    },
    "readthedocs": {
        "url": _githubusercontent_url(
            "readthedocs",
            "readthedocs.org",
            "main",
            "readthedocs/rtd_tests/fixtures/spec/v2/schema.json",
        ),
        "hook_config": {
            "name": "Validate ReadTheDocs Config",
            "description": (
                "Validate ReadTheDocs config against the schema provided by ReadTheDocs"
            ),
            "files": r"^\.readthedocs\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "renovate": {
        "url": "https://docs.renovatebot.com/renovate-schema.json",
        "hook_config": {
            "name": "Validate Renovate Config",
            "description": (
                "Validate Renovate config against the schema provided by "
                "Renovate (does not support renovate config in package.json)"
            ),
            "add_args": ["--regex-variant", "nonunicode"],
            "files": [
                r"renovate\.(json|json5)",
                r"\.(github|gitlab)/renovate\.(json|json5)",
                r"\.renovaterc",
                r"\.renovaterc\.(json|json5)",
            ],
        },
    },
    "snapcraft": {
        "url": _githubusercontent_url(
            "canonical", "snapcraft", "main", "schema/snapcraft.json"
        ),
        "hook_config": {
            "name": "Validate snapcraft files",
            "description": (
                "Validate snapcraft files against the schema provided by Canonical"
            ),
            "files": r"^(.+/)?snapcraft\.yaml$",
            "types": "yaml",
        },
    },
    "taskfile": {
        "url": "https://taskfile.dev/schema.json",
        "hook_config": {
            "name": "Validate Taskfile Config",
            "description": (
                "Validate Taskfile config against the schema provided by Task"
            ),
            "files": [
                r"Taskfile\.(yml|yaml)",
                r"taskfile\.(yml|yaml)",
                r"Taskfile\.dist\.(yml|yaml)",
                r"taskfile\.dist\.(yml|yaml)",
            ],
            "types": "yaml",
        },
    },
    "travis": {
        "url": "https://json.schemastore.org/travis",
        "hook_config": {
            "name": "Validate Travis Config",
            "files": r"^\.travis\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "woodpecker-ci": {
        "url": (
            "https://raw.githubusercontent.com/woodpecker-ci/woodpecker/main/pipeline"
            "/frontend/yaml/linter/schema/schema.json"
        ),
        "hook_config": {
            "name": "Validate Woodpecker Config",
            "files": [
                r"^\.woodpecker\.(yml|yaml)$",
                r"^\.woodpecker/.+\.(yml|yaml)$",
            ],
            "types": "yaml",
        },
    },
}



from __future__ import annotations

import os
import pathlib
import re
import typing as t
import urllib.parse
import urllib.request

import jsonschema

WINDOWS = os.name == "nt"

PROC_FD_PATH_PATTERN = re.compile(r"/proc/(self|\d+)/fd/\d+")

# this is a short list of schemes which will be recognized as being
# schemes at all; anything else will not even be reported as an
# unsupported scheme
KNOWN_URL_SCHEMES = [
    "",
    "ftp",
    "gopher",
    "http",
    "file",
    "https",
    "shttp",
    "rsync",
    "svn",
    "svn+ssh",
    "sftp",
    "nfs",
    "git",
    "git+ssh",
    "ws",
    "wss",
]


def is_url_ish(path: str) -> bool:
    r"""
    Returns true if the input path looks like a URL.

    NB: This needs to be done carefully to avoid mishandling of Windows paths
    starting with 'C:\' (and so forth) as URLs. urlparse from urllib will treat
    'C' as a scheme if asked to parse a Windows path.
    """
    if ":" not in path:
        return False
    scheme = path.split(":", 1)[0].lower()
    return scheme in KNOWN_URL_SCHEMES


def filename2path(filename: str) -> pathlib.Path:
    """
    Convert a filename which may be a local file URI to a pathlib.Path object

    This implementation was influenced strongly by how pip handles this problem:
      https://github.com/pypa/pip/blob/bf91a079791f2daf4339115fb39ce7d7e33a9312/src/pip/_internal/utils/urls.py#L26
    """
    if not filename.startswith("file://"):
        # for local paths, support use of `~`
        p = pathlib.Path(filename).expanduser()
    else:
        urlinfo = urllib.parse.urlsplit(filename)
        # local (vs UNC paths)
        is_local_path = urlinfo.netloc in (None, "", "localhost")

        if is_local_path:
            netloc = ""
        elif WINDOWS:
            netloc = "\\\\" + urlinfo.netloc
        else:
            netloc = urlinfo.netloc

        filename = urllib.request.url2pathname(netloc + urlinfo.path)

        # url2pathname on windows local paths can produce paths like
        #   /C:/Users/foo/...
        # the leading slash messes up a lot of logic for pathlib and similar functions
        # so strip the leading slash in this case
        if WINDOWS and is_local_path and filename.startswith("/"):
            filename = filename[1:]

        p = pathlib.Path(filename)

    # if passed a file descriptor object, do not try to resolve it
    # the resolution behavior when using zsh `<()` redirection seems to result in
    # an incorrect path being used
    if PROC_FD_PATH_PATTERN.fullmatch(filename):
        return p
    return p.resolve()


def iter_validation_error(
    err: jsonschema.ValidationError,
) -> t.Iterator[jsonschema.ValidationError]:
    if err.context:
        for e in err.context:
            yield e
            yield from iter_validation_error(e)



from __future__ import annotations

import typing as t

import ruamel.yaml


class Transform:
    def __init__(
        self,
        *,
        on_data: t.Callable[[list | dict], list | dict] | None = None,
    ) -> None:
        self.on_data = on_data

    def modify_yaml_implementation(self, implementation: ruamel.yaml.YAML) -> None:
        pass

    def __call__(self, data: list | dict) -> list | dict:
        if self.on_data is not None:
            return self.on_data(data)
        return data
